from typing import List, Callable, Dict
from pathlib import Path, PosixPath
import importlib.util
from . import ActionStore
from .loader import get_single_action_store
from .http import serve, report_error
from sys import argv, exit,path
from os import environ
import traceback


def get_funcs_from_file(infile: PosixPath) -> List[Callable]:
    if not infile.is_file():
        raise AssertionError(f"File {infile} does not exist")
    try:
        parents = str(infile.parent)
        filename = infile.name
        # print(f"Loading action store {filename} from {parents}")
        spec = importlib.util.spec_from_file_location(parents, infile.absolute())
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        ret = list()
        for o in dir(module):
            if o.startswith("_"):
                continue
            try:
                f = getattr(module, o)
                if not callable(f):
                    continue
                if not hasattr(f, "__module__") or f.__module__  != module.__name__:
                    continue
                ActionStore.validate_action(f)
                ret.append(f)
            except Exception as ex:
                print("Error loading action: "+module.__name__ + "."+ o, ex)
                pass
        return ret

    except Exception as e:
        print(traceback.format_exc())
        raise

def get_all_dir_files(dirpath: PosixPath) -> List[PosixPath]:
    dirpath = Path(dirpath)
    if not dirpath.exists():
        raise AssertionError(f"Directory {dirpath} does not exist")
    return list(dirpath.glob("**/*.py"))

if __name__ == "__main__":
    if len(argv) <2:
        print("Usage: python3 -m kubiya.bundle <paths>")
        paths = [Path(arg) for arg in argv[1:]]
        exit(1)
    funcs = {}
    files = []
    dirs = []
    try:
        for arg in argv[1:]:
            p = Path(arg)
            if p.is_dir():
                dirs.append(p)
                files.extend(get_all_dir_files(p))
                path.append(str(p))
            elif p.is_file():
                path.append(str(p.parent))
                files.append(p)
        
        for filename in files:
            filefuncs = get_funcs_from_file(filename)
            if filefuncs: 
                funcs[filename.stem] = filefuncs

        actionstore = get_single_action_store()
        serve(actionstore)
    except Exception as e:
        print(traceback.format_exc())
        report_error(e)


import json
from typing import List, Dict, NoReturn, Any, Iterable, Optional
from pydantic import validate_arguments, BaseModel
from jsonschema import validate, ValidationError
import inspect
from logging import getLogger
import functools
import pydantic

logger = getLogger(__name__)

class ActionStore:
    """handles kubiya actions and exeution"""

    _instances = list()

    @classmethod
    def register_action_store(cls, action_store: "ActionStore") -> NoReturn:
        cls._instances.append(action_store)
    
    def __init__(self, store_name: str, version: str = "testing") -> None:
        assert store_name != "", "store_name cannot be empty"
        self._registered_actions = {}
        self._action_metadata = {}
        self._registered_secrets = []
        self._name = store_name
        self._version = version
        self.__class__.register_action_store(self)

    @classmethod
    def validate_action(cls, action: callable) -> NoReturn:
        cls._validate_action_is_callable(action)
        cls._validate_action_signature(action)

    @classmethod
    def _validate_action_is_callable(cls, action: callable) -> NoReturn:
        assert callable(
            action
        ), f"{action} must be callable in order to be registered as an action"

    @classmethod
    def _validate_action_signature(cls, action: callable) -> NoReturn:
        sig = inspect.signature(action)
        mandatory_args = [
            nm
            for nm, par in sig.parameters.items()
            if par.default == inspect.Signature.empty
            and par.kind not in (par.VAR_KEYWORD, par.VAR_POSITIONAL)
        ]
        if len(mandatory_args) > 1:
            raise AssertionError(
                f"{action} must have at most 1 mandatory argument ({mandatory_args})"
            )

    def register_action(self, action_name: str, action: callable) -> NoReturn:
        self.validate_action(action)
        self._registered_actions[action_name] = action

    def uses_secrets(self, secret_names: Iterable[str]) -> NoReturn:
        self._registered_secrets.extend(secret_names)

    def get_registered_actions(self) -> List[str]:
        return list(self._registered_actions.keys())

    def get_registered_action_info(self, action_name: str) -> callable:
        assert (
            action_name in self._registered_actions
        ), f"`{action_name}` is not a registered action"
        return self._registered_actions[action_name]

    def get_registered_actions_info(self) -> List[callable]:
        return self._registered_actions.values()

    def get_registered_secrets(self) -> List[str]:
        return self._registered_secrets

    def execute_action(self, action_name: str, input: Optional[Dict]) -> Any:
        assert (
            action_name in self._registered_actions
        ), f"`{action_name}` is not a registered action"
        action = self._registered_actions[action_name]
        if input is None:
            return action()
        else:
            return action(input)

    def get_version(self) -> str:
        return self._version

    def get_name(self) -> str:
        return self._name

    @staticmethod
    def validate_args(func, input_schema=None):
        """if input type input_schema exists, validate against it"""
        def wrapper(*args, **kwargs):
            if input_schema:
                try:
                    first_arg = func.__code__.co_varnames[0]
                    validate(args[0], input_schema)
                except ValidationError as e:
                    return {"error": {"argument": first_arg, "field": e.schema.get("title"), "message": e.message, "expected_type" : e.schema.get("type")}}
            return func(*args, **kwargs)
        return wrapper

    
    def kubiya_action(self, validate_input=False) -> callable:
        """dict(type for typy in inspect...)"""
        def decorator(func: callable):
            action_data = {
                "action_name": func.__name__,
                "description": func.__doc__,
            }
            parameters = list(inspect.signature(func).parameters.values())
            return_signature = inspect.signature(func).return_annotation
            if return_signature == inspect.Signature.empty:
                output_schema = self.annotation_to_schema(
                    Any, "Output not specified")
            else:
                output_schema = self.annotation_to_schema(
                    return_signature, "Output")

            if len(parameters) > 1:
                raise RuntimeError(
                    f"{func.__name__} must have at most one argument to be registered as a kubiya action")
            if not parameters:
                input_schema = self.annotation_to_schema(
                    Any, "Input not specified")
            if parameters:
                model = parameters[0]
                input_schema = self.annotation_to_schema(
                    model.annotation, model.name)
            action_data["input_schema"] = input_schema
            action_data["output_schema"] = output_schema
            self._action_metadata[func.__name__] = action_data
            if not validate_input:
                self.register_action(func.__name__, func)
            else:
                if action_data.get("input_schema"):
                    wrapped = self.validate_args(func, input_schema=input_schema)
                    self.register_action(func.__name__, wrapped)
                    return wrapped
            self.register_action(func.__name__, func)
            return func
        return decorator


    @staticmethod
    def annotation_to_schema(annotation: Any, title: str) -> str:
        if type(annotation) == pydantic.main.ModelMetaclass:
            return json.loads(annotation.schema_json())
        elif annotation == inspect.Signature.empty:
            return json.loads(pydantic.schema_json_of(Any, title=title))
        else:
            return json.loads(pydantic.schema_json_of(annotation, title=title))



from kubiya import action_store
from kubiya.loader import load_action_store
from sys import argv, exit

def show_all():
    instances = action_store.ActionStore._instances
    if not instances:
        print("No action stores found")
        exit(1)
    if len(instances) > 1:
        print("Multiple action stores found: {instances}")
        exit(1)
    for instance in instances:
        print(f"{instance.get_name()}:")
        for action in instance.get_registered_actions():
            print(f"  {action}")

if __name__ == "__main__":    
    if len(argv) != 2:
        print("Usage: python3 -m kubiya.serve <action_store_file.py>")
        exit(1)
    
    load_action_store(argv[1])
    show_all()



from operator import ge
from fastapi import FastAPI
import uvicorn
import os

import kubiya
from .action_store import ActionStore
from typing import Optional, Dict, Any
from uvicorn import run
from pydantic import BaseModel
import traceback

class Request(BaseModel):
    action: Optional[str]
    input: Optional[Any]
    secrets: Optional[Dict]

async def execute_handler(request: Request, action_store: ActionStore) -> Any:
    try:
        if request.action == '__KUBIYA_DISCOVER__':
            return {
                "name": action_store.get_name(),
                "version": action_store.get_version(),
                "registered_actions": action_store.get_registered_actions(),
                "secrets": action_store.get_registered_secrets(),
                "kubiya_version": "python-sdk: " + kubiya.__version__,
                "actions_metadata": action_store._action_metadata,
            }

        if request.secrets:
            for secret, secret_value in request.secrets.items():
                print("Setting secret: ", secret)
                os.environ[secret] = secret_value
        else:
            print("No secrets defined")
        return action_store.execute_action(request.action, request.input)
    except Exception as e:
        return {
            "error": str(e),
            "stacktrace": traceback.format_exc(),
        }

def serve(action_store: ActionStore, filename=None):
    kubiya_server = FastAPI(openapi_url=None)

    @kubiya_server.post("/")
    async def root(request: Request):
        return await execute_handler(request=request, action_store=action_store)

    uvicorn.run(kubiya_server, host="0.0.0.0", port=8080)

def report_error(error: Exception):
    kubiya_server = FastAPI(openapi_url=None)

    @kubiya_server.post("/", status_code=500)
    async def root(request: Request):
        return {
            "error": {
                "message": str(error),
                "trace": traceback.format_exc(),
            }
        }

    uvicorn.run(kubiya_server, host="0.0.0.0", port=8080)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kubiya']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.79.0,<0.80.0',
 'jsonschema==4.17.3',
 'pydantic==1.10.2',
 'uvicorn>=0.18.2,<0.19.0']

setup_kwargs = {
    'name': 'kubiya',
    'version': '0.0.21',
    'description': 'Kubiya is a DevOps virtual assistant that captures and converts organizational knowledge into self-service workflows for a delightful business, IT and Cloud operations experience.',
    'long_description': 'None',
    'author': 'Kubiya',
    'author_email': 'info@kubiya.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.7.2,<3.12',
}


setup(**setup_kwargs)



from sys import exit
from pathlib import Path
import importlib.util
from . import action_store 

def load_action_store_file(file_path: str) -> str:
    file = Path(file_path).absolute()
    if not file.exists():
        raise AssertionError(f"File {file_path} does not exist")
    try:
        parents = str(file.parent)
        filename = file.name
        # print(f"Loading action store {filename} from {parents}")
        spec = importlib.util.spec_from_file_location(parents, file.absolute())
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module.__file__
    except Exception as e:
        print(f"Error loading actions store file: {e}")
        exit(1)

def load_action_store(store_name: str) -> str:
    try:
        return load_action_store_file(store_name)
    except Exception as e:
        print(f"Error loading store {store_name}: ({e})")
        exit(1)

def get_all_action_stores(filename=None):
    return action_store.ActionStore._instances

def get_single_action_store():
    action_stores = get_all_action_stores()
    if len(action_stores) == 0:
        raise AssertionError("No action stores found")
    if len(action_stores) > 1:
        raise AssertionError(f"Multiple action stores found: {action_stores}")
    return action_stores[0]


from kubiya import action_store
from kubiya.http import serve
from kubiya.loader import get_all_action_stores, load_action_store
from sys import argv, exit


def serve_all(filename=None):
    action_stores = get_all_action_stores()
    if not action_stores:
        print("No action stores found")
        exit(1)
    if len(action_stores) > 1:
        print("Multiple stores found: {instances}")
        exit(1)
    serve(action_stores[0], filename)

if __name__ == "__main__":    
    if len(argv) != 2:
        print("Usage: python3 -m kubiya.serve <action_store_file.py>")
        exit(1)
    
    store_file = load_action_store(argv[1])
    serve_all(store_file)
